"""Test BEL relations against Expression Atlas."""

from collections import namedtuple
from math import ceil
from typing import Dict, List, Set, Tuple

from flask import request
from sqlalchemy.sql import func

from ebel import Bel
from ebel.manager.rdbms.models.expression_atlas import (Experiment, FoldChange,
                                                        GroupComparison)
from ebel.web.api.ebel.v1 import _get_pagination

Relation = namedtuple(
    "Relation",
    ["sub", "sub_bel", "rel_rid", "evidence", "rel", "pmid", "obj", "obj_bel"],
)
ComparisonGroupValue = namedtuple("CompGroupFoldChange", ["group_comparison_id", "log2foldchange", "p_value"])


def get_fold_changes(
    session,
    gene_name: str,
    l2fc_threshold: float,
    pv_threshold: float,
    experiment_names: List[str],
) -> Tuple[List[ComparisonGroupValue], List[ComparisonGroupValue]]:
    """Get fold change values for a given set of passed parameters.

    Parameters
    ----------
    session
        SQL session.
    gene_name : str
        Gene name.
    l2fc_threshold : float
        Minimum log2 fold change threshold.
    pv_threshold : float
        P-value threshold. Recommended value: 0.05
    experiment_names : List[str]
        List of experiment names.

    Returns
    -------
    Tuple.
    """
    query = (
        session.query(
            FoldChange.group_comparison_id,
            func.avg(FoldChange.log2foldchange),
            FoldChange.p_value,
        )
        .filter_by(gene_name=gene_name)
        .group_by(FoldChange.group_comparison_id, FoldChange.gene_id)
    )
    if experiment_names:
        query = query.join(GroupComparison).join(Experiment).filter(Experiment.name.in_(experiment_names))

    query_up = query.filter(FoldChange.log2foldchange >= l2fc_threshold, FoldChange.p_value <= pv_threshold)
    query_down = query.filter(
        FoldChange.log2foldchange <= -1 * l2fc_threshold,
        FoldChange.p_value <= pv_threshold,
    )

    up_all = [ComparisonGroupValue(*x) for x in query_up.all()]
    down_all = [ComparisonGroupValue(*x) for x in query_down.all()]
    return down_all, up_all


def validate_bel_against_experiment() -> dict:
    """Check whether BEL edges match experimental results."""
    pagination = _get_pagination()

    subject_class = request.args.get("subject_class", "genetic_flow")
    subject_name = request.args.get("subject_hgnc_symbol")
    object_class = request.args.get("object_class", "genetic_flow")
    object_name = request.args.get("object_hgnc_symbol")
    subject_l2fc_threshold = float(request.args.get("subject_log2foldchange_threshold", "2"))
    subject_pv_threshold = float(request.args.get("subject_p_value_threshold", "0.05"))
    object_l2fc_threshold = float(request.args.get("object_log2foldchange_threshold", "2"))
    object_pv_threshold = float(request.args.get("object_p_value_threshold", "0.05"))

    experiment_names = [x.strip() for x in request.args.get("experiment_names", "").split(",") if x.strip()]

    subject_sql = f" and name = '{subject_name}'" if subject_name else ""
    object_sql = f" and name = '{object_name}'" if object_name else ""

    sql_basis = f"""from (
        match {{class:{subject_class}, as:sub,
        where:(namespace = 'HGNC' {subject_sql})}}
        .outE('directly_increases', 'increases','decreases',
        'directly_decreases', 'positive_correlation', 'negative_correlation'){{as:rel}}
        .inV(){{class:{object_class}, as:obj,
        where:(namespace = 'HGNC' {object_sql})}}
        return
            sub.name as sub,
            sub.bel as sub_bel,
            rel.@rid.asString() as rel_rid,
            rel.evidence as evidence,
            rel.@class.asString() as rel,
            rel.pmid as pmid,
            obj.name as obj,
            obj.bel as obj_bel)
        where sub!=obj"""

    bel = Bel()
    session = bel.session
    number_of_results = bel.execute("Select count(*) " + sql_basis)[0].oRecordData["count"]
    sql = "Select * " + sql_basis + f" limit {pagination.page_size} skip {pagination.skip}"
    relations: List[Relation] = [Relation(**x.oRecordData) for x in bel.execute(sql)]

    dea_results = {}
    gc_ids = set()

    for gene_name_sub, gene_name_obj in {(x.sub, x.obj) for x in relations}:
        sub_down, sub_up = get_fold_changes(
            session,
            gene_name_sub,
            subject_l2fc_threshold,
            subject_pv_threshold,
            experiment_names,
        )
        sub_down_gc_set = {x.group_comparison_id for x in sub_down}
        sub_up_gc_set = {x.group_comparison_id for x in sub_up}

        if gene_name_sub not in dea_results:
            dea_results[gene_name_sub] = {}

        obj_down, obj_up = get_fold_changes(
            session,
            gene_name_obj,
            object_l2fc_threshold,
            object_pv_threshold,
            experiment_names,
        )
        obj_down_gc_set = {x.group_comparison_id for x in obj_down}
        obj_up_gc_set = {x.group_comparison_id for x in obj_up}

        res = dea_results[gene_name_sub][gene_name_obj] = {}

        # opposite direction
        sub_up_obj_down_values = get_values_for_intersection(sub_up, obj_down, sub_up_gc_set, obj_down_gc_set)
        gc_ids.update(set(sub_up_obj_down_values.keys()))
        res["sub_up_obj_down"] = sub_up_obj_down_values

        sub_down_obj_up_values = get_values_for_intersection(sub_down, obj_up, sub_down_gc_set, obj_up_gc_set)
        gc_ids.update(set(sub_down_obj_up_values.keys()))
        res["sub_down_obj_up"] = sub_down_obj_up_values

        # same direction
        sub_up_obj_up_values = get_values_for_intersection(sub_up, obj_up, sub_up_gc_set, obj_up_gc_set)
        gc_ids.update(set(sub_up_obj_up_values.keys()))
        res["sub_up_obj_up"] = sub_up_obj_up_values

        sub_down_obj_down_values = get_values_for_intersection(sub_down, obj_down, sub_down_gc_set, obj_down_gc_set)
        gc_ids.update(set(sub_down_obj_down_values.keys()))
        res["sub_down_obj_down"] = sub_down_obj_down_values

    return {
        "dea": dea_results,
        "group_comparisons": get_group_experiment_by_gc_ids(session, gc_ids),
        "bel_relations": get_bel_relations(relations, dea_results),
        "page": pagination.page,
        "page_size": pagination.page_size,
        "number_of_results": number_of_results,
        "pages": ceil(number_of_results / pagination.page_size),
    }


def get_bel_relations(relations: List[Relation], dea_results) -> List[dict]:
    """Create a list of dictionaries describing BEL edges that are associated with the gene symbols for some DEA.

    Parameters
    ----------
    relations : List[Relation]
    dea_results : dict

    Returns
    -------
    List of BEL relation metadata.
    """
    results = []
    for relation in relations:
        dea = dea_results[relation.sub][relation.obj]
        rel_dict = relation._asdict()
        same_direction = list(dea["sub_up_obj_up"].keys()) + list(dea["sub_down_obj_down"].keys())
        opposite_direction = list(dea["sub_up_obj_down"].keys()) + list(dea["sub_down_obj_up"].keys())
        if relation.rel in ["directly_increases", "increases", "positive_correlation"]:
            rel_dict["supported_by"] = same_direction
            rel_dict["in_contradiction_to"] = opposite_direction
        elif relation.rel in [
            "directly_decreases",
            "decreases",
            "negative_correlation",
        ]:
            rel_dict["supported_by"] = opposite_direction
            rel_dict["in_contradiction_to"] = same_direction
        results.append(rel_dict)
    return results


def get_group_experiment_by_gc_ids(session, group_comparison_ids: Set[int]) -> dict:
    """Return a dictionary of experimental metadata for a given set of group comparison IDs.

    Parameters
    ----------
    session
        SQL session.
    group_comparison_ids : Set[int]
        Group comparison IDs.

    Returns
    -------
    Dictionary with group comparison ID as the keys and the associated experiment metadata as a value dictionary.
    """
    results = {}
    for group_comparison_id in group_comparison_ids:
        gc: GroupComparison = session.query(GroupComparison).filter_by(id=group_comparison_id).first()
        results[group_comparison_id] = {
            "experiment_id": gc.experiment.id,
            "experiment": gc.experiment.name,
            "experiment_title": gc.experiment.title,
            "group_comparison": gc.group_comparison,
            "group_comparison_title": gc.name,
        }
    return results


def get_values_for_intersection(
    sub: List[ComparisonGroupValue],
    obj: List[ComparisonGroupValue],
    sub_gc_set: Set[int],
    obj_gc_set: Set[int],
) -> Dict[int, dict]:
    """Get the log2 fold changes and p values for the intersection of a given subject and object.

    Parameters
    ----------
    sub : List[ComparisonGroupValue]
    obj : List[ComparisonGroupValue]
    sub_gc_set : Set[int]
    obj_gc_set : Set[int]

    Returns
    -------
    Dictionary with group comparison ID as keys and the log2 fold changes + p values as a value dictionary.
    """
    sub_obj_gc_ids = sub_gc_set & obj_gc_set
    gc_sub_dict = {x.group_comparison_id: {"l2fc": x.log2foldchange, "pv": x.p_value} for x in sub}
    gc_obj_dict = {x.group_comparison_id: {"l2fc": x.log2foldchange, "pv": x.p_value} for x in obj}

    results = {}
    for sub_obj_gc_id in sub_obj_gc_ids:
        results[sub_obj_gc_id] = {
            "sub": gc_sub_dict[sub_obj_gc_id],
            "obj": gc_obj_dict[sub_obj_gc_id],
        }
    return results



"""Init API unit tests."""

import json
from typing import Union

import pytest

from ebel.web.app import application


@pytest.fixture(scope="module")
def client():
    with application.app.test_client() as client:
        yield client


def format_response_data(response) -> Union[dict, list]:
    """Used to query to test client and retrieve the data."""
    assert response.status_code == 200
    return json.loads(response.data)



"""DisGeNet API methods."""
from flask import request

from ebel.manager.rdbms.models import disgenet
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_paginated_ebel_query_result,
                                  _get_paginated_query_result,
                                  _get_terms_from_model_starts_with)


def get_sources():
    """Return all sources."""
    d = disgenet.DisgenetSource
    return {x.source: x.id for x in RDBMS.get_session().query(d.source, d.id).all()}


def get_disease_name_starts_with():
    """Get entry based on beginning of disease name."""
    return _get_terms_from_model_starts_with(
        form_field="disease_name", sa_column=disgenet.DisgenetDisease.disease_name
    )


def get_gene_symbol_starts_with():
    """Get entry based on beginning of symbol."""
    return _get_terms_from_model_starts_with(
        form_field="gene_symbol", sa_column=disgenet.DisgenetGeneSymbol.gene_symbol
    )


def get_gene_disease_pmid_associations():
    """Get list of PMIDs with disease genes defined."""
    gene_cols = ["gene_id", "disease_id", "pmid"]
    gene_paras = {k: v for k, v in request.args.items() if k in gene_cols and v}
    query = RDBMS.get_session().query(disgenet.DisgenetGene).filter_by(**gene_paras)
    gene_symbol = request.args.get("gene_symbol")
    if gene_symbol:
        gs = disgenet.DisgenetGeneSymbol
        query = query.join(gs).filter(gs.gene_symbol.like(gene_symbol))
    disease_name = request.args.get("disease_name")
    if disease_name:
        dd = disgenet.DisgenetDisease
        query = query.join(dd).filter(dd.disease_name.like(disease_name))
    source = request.args.get("source")
    if source:
        so = disgenet.DisgenetSource
        query = query.join(so).filter(so.source.like(source))

    return _get_paginated_query_result(query)


def get_variant_disease_pmid_associations():
    """Get list of PMIDs with disease variants defined."""
    cols = [
        "snp_id",
        "chromosome",
        "position",
        "disease_id",
        "score",
        "pmid",
        "source_id",
    ]
    paras = {k: v for k, v in request.args.items() if k in cols and v}
    query = RDBMS.get_session().query(disgenet.DisgenetVariant).filter_by(**paras)
    disease_name = request.args.get("disease_name")
    if disease_name:
        dd = disgenet.DisgenetDisease
        query = query.join(dd).filter(dd.disease_name.like(disease_name))
    source = request.args.get("source")
    if source:
        so = disgenet.DisgenetSource
        query = query.join(so).filter(so.source.like(source))

    return _get_paginated_query_result(query)


def get_ebel_has_snp_disgenet():
    """Get SNP edge add by e(BE:L) via DiGeNet."""
    edge_class = "has_snp_disgenet"

    allowed_relations = [
        "has_snp_disgenet",
        "has_mapped_snp_dgn",
        "has_downstream_snp_dgn",
        "has_upstream_snp_dgn",
    ]
    relation = request.args.get("relation")

    where = {
        "@class": relation if relation in allowed_relations else None,
        "out.name": request.args.get("name"),
        "out.namespace": request.args.get("namespace"),
        "in.rs_number": request.args.get("rs_number"),
    }
    # non-edge parameter
    # edge parameter
    edge_properties = ["disease_name", "source", "pmid"]
    where.update({k: v for k, v in request.args.items() if k in edge_properties})

    sql = f"""SELECT
        @rid.asString() as rid,
        @class as relation,
        in.rs_number as snp_rs_number,
        in.@rid.asString() as snp_rid,
        score,
        disease_name,
        source,
        set(pmid) as pmids,
        out.@rid.asString() as gene_rid,
        out.name as gene_symbol FROM {edge_class}"""
    if where:
        sql += " WHERE " + " AND ".join([f'{k} like "{v}"' for k, v in where.items() if v])
    sql += " GROUP BY @class, in.rs_number, score, disease_name, source, out.name"

    return _get_paginated_ebel_query_result(sql)



"""Protein Atlas RDBMS model definition."""
from sqlalchemy import Column, Integer, Numeric, String, Text
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class ProteinAtlasNormalTissue(Base):
    """Class definition for the protein_atlas_normal_tissue table."""

    __tablename__ = "protein_atlas_normal_tissue"
    id = Column(Integer, primary_key=True)

    gene = Column(String(100), index=True)
    gene_name = Column(String(100))
    tissue = Column(String(100))
    cell_type = Column(String(100))
    level = Column(String(100), index=True)
    reliability = Column(String(100), index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "gene": self.gene,
            "gene_name": self.gene_name,
            "tissue": self.tissue,
            "cell_type": self.cell_type,
            "level": self.level,
            "reliability": self.reliability,
        }


class ProteinAtlasSubcellularLocation(Base):
    """Class definition for the protein_atlas_subcellular_location table."""

    __tablename__ = "protein_atlas_subcellular_location"
    id = Column(Integer, primary_key=True)

    gene = Column(String(100))
    gene_name = Column(String(100))
    reliability = Column(String(100))
    main_location = Column(String(100))
    additional_location = Column(String(100))
    extracellular_location = Column(String(100))
    enhanced = Column(String(100))
    supported = Column(String(100))
    approved = Column(String(100))
    uncertain = Column(String(100))
    single_cell_variation_intensity = Column(String(100))
    single_cell_variation_spatial = Column(String(100))
    cell_cycle_dependency = Column(Text)
    go_id = Column(Text)

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "gene": self.gene,
            "gene_name": self.gene_name,
            "reliability": self.reliability,
            "main_location": self.main_location,
            "additional_location": self.additional_location,
            "extracellular_location": self.extracellular_location,
            "enhanced": self.enhanced,
            "supported": self.supported,
            "approved": self.approved,
            "uncertain": self.uncertain,
            "single_cell_variation_intensity": self.single_cell_variation_intensity,
            "single_cell_variation_spatial": self.single_cell_variation_spatial,
            "cell_cycle_dependency": self.cell_cycle_dependency,
            "go_id": self.go_id,
        }


class ProteinAtlasRnaTissueConsensus(Base):
    """Class definition for the protein_atlas_rna_tissue_consensus table."""

    __tablename__ = "protein_atlas_rna_tissue_consensus"
    id = Column(Integer, primary_key=True)

    gene = Column(String(100), index=True)
    gene_name = Column(String(100), index=True)
    tissue = Column(String(100), index=True)
    n_tpm = Column(Numeric(8, 1))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "gene": self.gene,
            "gene_name": self.gene_name,
            "tissue": self.tissue,
            "n_tpm": self.nx,
        }


class ProteinAtlasRnaBrainGtex(Base):
    """Class definition for the protein_atlas_rna_brain_gtex table."""

    __tablename__ = "protein_atlas_rna_brain_gtex"
    id = Column(Integer, primary_key=True)

    gene = Column(String(100), index=True)
    gene_name = Column(String(100), index=True)
    brain_region = Column(String(100), index=True)
    tpm = Column(Numeric(8, 1))
    p_tpm = Column(Numeric(8, 1))
    n_tpm = Column(Numeric(8, 1))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "gene": self.gene,
            "gene_name": self.gene_name,
            "brain_region": self.brain_region,
            "tpm": self.tpm,
            "p_tpm": self.p_tpm,
            "n_tpm": self.nx,
        }


class ProteinAtlasRnaBrainFantom(Base):
    """Class definition for the protein_atlas_rna_brain_fantom table."""

    __tablename__ = "protein_atlas_rna_brain_fantom"
    id = Column(Integer, primary_key=True)

    gene = Column(String(100))
    gene_name = Column(String(100))
    brain_region = Column(String(100))
    tags_per_million = Column(String(100))
    scaled_tags_per_million = Column(String(100))
    n_tpm = Column(String(100))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "gene": self.gene,
            "gene_name": self.gene_name,
            "brain_region": self.brain_region,
            "tags_per_million": self.tags_per_million,
            "scaled_tags_per_million": self.scaled_tags_per_million,
            "n_tpm": self.nx,
        }


class ProteinAtlasRnaMouseBrainAllen(Base):
    """Class definition for the protein_atlas_rna_mouse_brain_allen table."""

    __tablename__ = "protein_atlas_rna_mouse_brain_allen"
    id = Column(Integer, primary_key=True)

    gene = Column(String(100))
    gene_name = Column(String(100))
    brain_region = Column(String(100))
    expression_energy = Column(Numeric(8, 1))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "gene": self.gene,
            "gene_name": self.gene_name,
            "brain_region": self.brain_region,
            "expression_energy": self.expression_energy,
        }



"""NCBI RDBMS model definition."""
from sqlalchemy import Column, ForeignKey, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from . import object_as_dict

Base = declarative_base()


class NcbiGeneInfo(Base):
    """Class definition for the ncbi_gene_info table."""

    __tablename__ = "ncbi_gene_info"
    gene_id = Column(Integer, primary_key=True)

    tax_id = Column(Integer, index=True)
    symbol = Column(String(100), index=True)
    type_of_gene = Column(String(100), index=True)
    locus_tag = Column(String(100))
    chromosome = Column(String(100))
    map_location = Column(String(100))
    description_id = Column(Integer, ForeignKey("ncbi_gene_info_description.id"))
    description = relationship("NcbiGeneInfoDescription", foreign_keys=[description_id])
    xrefs = relationship("NcbiGeneInfoXref", back_populates="gene")
    mims = relationship("NcbiGeneMim", foreign_keys="NcbiGeneMim.gene_id", back_populates="gene")
    orthologs = relationship(
        "NcbiGeneOrtholog",
        foreign_keys="NcbiGeneOrtholog.gene_id",
        back_populates="gene",
    )
    ensembl_ids = relationship("NcbiGeneEnsembl", back_populates="genes")
    gene_ids_right = relationship("NcbiGeneOnRight", foreign_keys="NcbiGeneOnRight.gene_id", back_populates="gene")
    gene_ids_left = relationship("NcbiGeneOnLeft", foreign_keys="NcbiGeneOnLeft.gene_id", back_populates="gene")
    gene_ids_overlapping = relationship(
        "NcbiGeneOverlapping",
        foreign_keys="NcbiGeneOverlapping.gene_id",
        back_populates="gene",
    )

    def as_dict(self):
        """Convert object values to dictionary."""
        rdict = object_as_dict(self, ["description_id"])
        rdict.update(
            {
                "xrefs": [{"db": x.db, "dbid": x.dbid} for x in self.xrefs],
                "mims": [x.mim_number for x in self.mims],
                "description": self.description.description,
                "ensembls": [x.ensembl_gene_identifier for x in self.ensembl_ids],
                "orthologs": [x.other_gene_id for x in self.orthologs],
                "gene_ids_right": [x.gene_id_on_right for x in self.gene_ids_right],
                "gene_ids_left": [x.gene_id_on_left for x in self.gene_ids_left],
                "gene_ids_overlapping": [x.overlapping_gene_id for x in self.gene_ids_overlapping],
            }
        )
        return rdict


class NcbiGeneInfoDescription(Base):
    """Class definition for the ncbi_gene_info_description table."""

    __tablename__ = "ncbi_gene_info_description"
    id = Column(Integer, primary_key=True, autoincrement=True)
    description = Column(Text)


class NcbiGeneOnRight(Base):
    """Class definition for the ncbi_gene_on_right table."""

    __tablename__ = "ncbi_gene_on_right"
    id = Column(Integer, primary_key=True, autoincrement=True)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    gene_id_on_right = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGeneOnLeft(Base):
    """Class definition for the ncbi_gene_on_left table."""

    __tablename__ = "ncbi_gene_on_left"
    id = Column(Integer, primary_key=True, autoincrement=True)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    gene_id_on_left = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGeneOverlapping(Base):
    """Class definition for the ncbi_gene_overlapping table."""

    __tablename__ = "ncbi_gene_overlapping"
    id = Column(Integer, primary_key=True, autoincrement=True)
    gene_id = gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    overlapping_gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGeneOrtholog(Base):
    """Class definition for the ncbi_gene_ortholog table."""

    __tablename__ = "ncbi_gene_ortholog"
    id = Column(Integer, primary_key=True, autoincrement=True)
    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    other_tax_id = Column(Integer, index=True)
    other_gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGenePubmed(Base):
    """Class definition for the ncbi_gene_pubmed table."""

    __tablename__ = "ncbi_gene_pubmed"
    id = Column(Integer, primary_key=True)

    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    pub_med_id = Column(Integer)


class NcbiGeneInfoXref(Base):
    """Class definition for the ncbi_gene_info_xref table."""

    __tablename__ = "ncbi_gene_info_xref"
    id = Column(Integer, primary_key=True, autoincrement=True)

    db = Column(String(100), index=True)
    dbid = Column(String(100), index=True)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))

    gene = relationship("NcbiGeneInfo", back_populates="xrefs")


class NcbiGeneMim(Base):
    """Class definition for the ncbi_gene_mim table."""

    __tablename__ = "ncbi_gene_mim"
    id = Column(Integer, primary_key=True, autoincrement=True)

    mim_number = Column(Integer)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    type = Column(String(100))
    source = Column(String(100))
    med_gen_cui = Column(String(100), index=True)
    comment = Column(String(100))

    gene = relationship("NcbiGeneInfo", back_populates="mims")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "mim_number": self.mim_number,
            "gene_id": self.gene_id,
            "type": self.type,
            "source": self.source,
            "med_gen_cui": self.med_gen_cui,
            "comment": self.comment,
        }


class NcbiGeneEnsembl(Base):
    """Class definition for the ncbi_gene_ensembl table."""

    __tablename__ = "ncbi_gene_ensembl"
    id = Column(Integer, primary_key=True, autoincrement=True)

    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    ensembl_gene_identifier = Column(String(100))
    rna_nucleotide_accession_version = Column(String(100))
    ensembl_rna_identifier = Column(String(100))
    protein_accession_version = Column(String(100))
    ensembl_protein_identifier = Column(String(100))

    genes = relationship("NcbiGeneInfo", back_populates="ensembl_ids")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "tax_id": self.tax_id,
            "gene_id": self.gene_id,
            "ensembl_gene_identifier": self.ensembl_gene_identifier,
            "rna_nucleotide_accession_version": self.rna_nucleotide_accession_version,
            "ensembl_rna_identifier": self.ensembl_rna_identifier,
            "protein_accession_version": self.protein_accession_version,
            "ensembl_protein_identifier": self.ensembl_protein_identifier,
        }


class NcbiGeneGo(Base):
    """Class definition for the ncbi_gene_go table."""

    __tablename__ = "ncbi_gene_go"
    id = Column(Integer, primary_key=True, autoincrement=True)

    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey("ncbi_gene_info.gene_id"))
    go_id = Column(String(100), index=True)
    evidence = Column(String(10))
    qualifier = Column(String(100))
    go_term = Column(String(255))
    category = Column(String(10))

    pmids = relationship("NcbiGeneGoPmid", back_populates="gos")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "tax_id": self.tax_id,
            "gene_id": self.gene_id,
            "go_id": self.go_id,
            "evidence": self.evidence,
            "qualifier": self.qualifier,
            "go_term": self.go_term,
            "pmids": [x.pmid for x in self.pmids],
            "category": self.category,
        }


class NcbiGeneGoPmid(Base):
    """Class definition for the ncbi_gene_go_pmid table."""

    __tablename__ = "ncbi_gene_go_pmid"
    id = Column(Integer, primary_key=True, autoincrement=True)

    ncbi_gene_go_id = Column(Integer, ForeignKey("ncbi_gene_go.id"))
    pmid = Column(Integer)

    gos = relationship("NcbiGeneGo", back_populates="pmids")


class NcbiMedGenName(Base):
    """Class definition for the ncbi_medgen_name table."""

    __tablename__ = "ncbi_medgen_name"
    id = Column(Integer, primary_key=True, autoincrement=True)

    cui = Column(String(100))
    name = Column(Text)
    source = Column(String(100))
    suppress = Column(String(1))
    pmids = relationship("NcbiMedGenPmid", back_populates="med_gen_name")

    def as_dict(self):
        """Convert object values to dictionary."""
        rdit = object_as_dict(self, exclude=["id"])
        rdit.update({"pmids": [x.pmid for x in self.pmids]})
        return rdit


class NcbiMedGenPmid(Base):
    """Class definition for the ncbi_medgen_pmid table."""

    __tablename__ = "ncbi_medgen_pmid"
    id = Column(Integer, primary_key=True, autoincrement=True)

    ncbi_medgen_name_id = Column(Integer, ForeignKey("ncbi_medgen_name.id"))
    pmid = Column(Integer, index=True)

    med_gen_name = relationship("NcbiMedGenName", back_populates="pmids")

    def as_dict(self):
        """Convert object values to dictionary."""
        rdit = object_as_dict(self, exclude=["id"])
        return rdit



"""Pathway Commons API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestPathwayCommons:
    def test_get_pathway_commons(self, client):
        response = client.get(
            "api/v1/pathway_commons?participant_a=CSF3&interaction_type=controls-expression-of&participant_b=CD33&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "id": 804357,
            "interaction_type": "controls-expression-of",
            "participant_a": "CSF3",
            "participant_b": "CD33",
            "pathway_names": [],
            "pmids": [12627849],
            "sources": ["CTD"],
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_by_gene_symbol(self, client):
        response = client.get(
            "api/v1/pathway_commons/by_gene_symbol?gene_symbol=CD33&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_cols = (
            "id",
            "interaction_type",
            "participant_a",
            "participant_b",
            "pathway_names",
            "pmids",
            "sources",
        )
        assert output[NUM_RESULTS] >= 19
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["participant_b"] == "CD33"

    def test_get_by_pathway_name(self, client):
        response = client.get(
            "api/v1/pathway_commons/by_pathway_name?pathway_name=Activation%20and%20oligomerization%20of%20BAK%20protein&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_result = {
            "id": 69305,
            "interaction_type": "in-complex-with",
            "participant_a": "BAK1",
            "participant_b": "BID",
            "pathway_names": ["Activation and oligomerization of BAK protein"],
            "pmids": [],
            "sources": ["Reactome"],
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_result

    def test_get_by_pmid(self, client):
        response = client.get(
            "api/v1/pathway_commons/by_pmid?pmid=12627849&page_size=10&page=1", content_type="application/json"
        )
        output = format_response_data(response)
        expected_cols = (
            "id",
            "interaction_type",
            "participant_a",
            "participant_b",
            "pathway_names",
            "pmids",
            "sources",
        )
        assert output[NUM_RESULTS] >= 6
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert 12627849 in hit["pmids"]

    def test_get_pathway_name_starts_with(self, client):
        response = client.get(
            "api/v1/pathway_commons/pathway_name/starts_with?pathway_name=Activation%20and%20oligomerization%20of%20BAK&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        hit = output[RESULTS]

        assert isinstance(hit, dict)
        assert "Activation and oligomerization of BAK protein" in hit
        assert isinstance(hit["Activation and oligomerization of BAK protein"], int)



"""IntAct module."""
import logging
import zipfile
from typing import Dict

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.biodbs.uniprot import UniProt
from ebel.manager.orientdb.constants import INTACT
from ebel.manager.rdbms.models import intact, uniprot
from ebel.tools import get_file_path

logger = logging.getLogger(__name__)


class IntAct(odb_meta.Graph):
    """IntAct."""

    def __init__(self, client: OrientDB = None, condition_keyword="Alzheimer"):
        """Init IntAct."""
        self.condition_keyword = condition_keyword
        self.client = client
        self.biodb_name = INTACT
        self.urls = {self.biodb_name: urls.INTACT}
        self.file_path = get_file_path(urls.INTACT, self.biodb_name)
        super().__init__(
            edges=odb_structure.intact_edges,
            indices=odb_structure.intact_indices,
            tables_base=intact.Base,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    @staticmethod
    def _intact_list_to_dict(string_list: str) -> dict:
        dict_values = dict()
        split_values = [x.split(":", 1) for x in string_list.split("|")]
        for values in split_values:
            if len(values) == 2:
                key, val = values[0], values[1]

                if key in dict_values:
                    dict_values[key].append(val)
                else:
                    dict_values[key] = [val]

        return dict_values

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class."""
        self.recreate_tables()

        logger.info("Reading and formatting IntAct data.")

        zf = zipfile.ZipFile(self.file_path)

        usecols = {
            "#ID(s) interactor A": "int_a_uniprot_id",
            "ID(s) interactor B": "int_b_uniprot_id",
            "Publication Identifier(s)": "pmid",
            "Interaction type(s)": "it",
            "Interaction identifier(s)": "interaction_ids",
            "Confidence value(s)": "confidence_value",
            "Interaction detection method(s)": "dm",
        }

        df = pd.read_csv(zf.open("intact.txt"), sep="\t", usecols=usecols.keys())
        df.rename(columns=usecols, inplace=True)
        regex_accession = r"uniprotkb:([OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2})"
        df.int_a_uniprot_id = df.int_a_uniprot_id.str.extract(regex_accession)[0]
        df.int_b_uniprot_id = df.int_b_uniprot_id.str.extract(regex_accession)[0]
        df = df[(pd.notnull(df.int_a_uniprot_id) & pd.notnull(df.int_b_uniprot_id))]
        regex_detection_method = r"psi-mi:\"MI:0*(?P<detection_method_psimi_id>\d+)\"\((?P<detection_method>[^)]+)\)"
        df = df.join(df.dm.str.extract(regex_detection_method), how="left")
        df.drop(columns=["dm"], inplace=True)
        df.pmid = df.pmid.str.extract(r"pubmed:(\d+)")
        regex_interaction_type = r"psi-mi:\"MI:0*(?P<interaction_type_psimi_id>\d+)\"\((?P<interaction_type>[^)]+)\)"
        df = df.join(df.it.str.extract(regex_interaction_type), how="left")
        df.drop(columns=["it"], inplace=True)
        df.confidence_value = df.confidence_value.str.extract(r"intact-miscore:(\d+(\.\d+)?)")[0]
        df.index += 1
        df.index.rename("id", inplace=True)

        df.to_sql(intact.Intact.__tablename__, self.engine, if_exists="append")

        self.session.commit()

        return {self.biodb_name: df.shape[0]}

    def get_create_rid_by_uniprot(self, uniprot_accession: str, uniprot_rid_dict: dict) -> str:
        """Create or get rID entry for a given UniProt ID.

        Parameters
        ----------
        uniprot_accession: str
            UniProt accession number.
        uniprot_rid_dict: dict
            Entry parameters matching those of the desired rID entry.

        Returns
        -------
        str
            UniProt accession ID.
        """
        if uniprot_accession not in uniprot_rid_dict:
            nn = self.get_namespace_name_by_uniprot(uniprot_accession)
            if nn:
                namespace, name = nn
                value_dict = {
                    "name": name,
                    "namespace": namespace,
                    "pure": True,
                    "bel": f'p({namespace}:"{name}")',
                    "uniprot": uniprot_accession,
                }
                uniprot_rid_dict[uniprot_accession] = self.get_create_rid("protein", value_dict, check_for="bel")
        return uniprot_rid_dict.get(uniprot_accession)

    def get_namespace_name_by_uniprot(self, uniprot_accession: str) -> tuple:
        """Get the namespace of a given UniProt ID.

        Parameters
        ----------
        uniprot_accession: str
            UniProt accession number.

        Returns
        -------
        tuple
            namespace, value
        """
        return_value = ()
        sql = f"""Select s.symbol, u.taxid from uniprot u inner join uniprot_gene_symbol s
                  on (u.id=s.uniprot_id) where u.accession='{uniprot_accession}' limit 1"""
        result = self.engine.execute(sql).fetchone()
        taxid_to_namespace = {9606: "HGNC", 10090: "MGI", 10116: "RGD"}
        if result:
            name, taxid = result
            namespace = taxid_to_namespace.get(taxid, "UNIPROT")
            return_value = (namespace, name)
        else:
            if self.session.query(uniprot.Uniprot).filter(uniprot.Uniprot.accession == uniprot_accession).first():
                return_value = ("UNIPROT", uniprot_accession)
        return return_value

    def update_interactions(self) -> int:
        """Update intact interactions to graph."""
        uniprot = UniProt(self.client)
        uniprot.update()

        uniprot_rid_dict = uniprot.get_pure_uniprot_rid_dict_in_bel_context()

        sql_temp = """SELECT
            int_a_uniprot_id,
            int_b_uniprot_id,
            pmid,
            interaction_ids,
            interaction_type,
            interaction_type_psimi_id,
            detection_method,
            detection_method_psimi_id,
            confidence_value
        FROM
            intact
        WHERE
            int_a_uniprot_id = '{uniprot_accession}' or int_b_uniprot_id = '{uniprot_accession}'
        GROUP BY
            int_a_uniprot_id,
            int_b_uniprot_id,
            pmid,
            interaction_ids,
            interaction_type,
            interaction_type_psimi_id,
            detection_method,
            detection_method_psimi_id,
            confidence_value"""

        updated = 0

        uniprot_accessions = tuple(uniprot_rid_dict.keys())

        for uniprot_accession in tqdm(uniprot_accessions, desc="Update IntAct interactions"):
            sql = sql_temp.format(uniprot_accession=uniprot_accession)
            result = self.engine.execute(sql)

            for (
                up_a,
                up_b,
                pmid,
                ia_ids,
                ia_type,
                ia_type_id,
                d_method,
                d_method_id,
                c_value,
            ) in result.fetchall():
                from_rid = self.get_create_rid_by_uniprot(up_a, uniprot_rid_dict)
                to_rid = self.get_create_rid_by_uniprot(up_b, uniprot_rid_dict)

                if from_rid and to_rid:
                    value_dict = {
                        "interaction_type": ia_type,
                        "interaction_type_psimi_id": ia_type_id,
                        "detection_method": d_method,
                        "detection_method_psimi_id": d_method_id,
                        "interaction_ids": dict([x.split(":", 1) for x in ia_ids.split("|")]),
                        "confidence_value": float(c_value),
                        "pmid": pmid,
                    }

                    self.create_edge(
                        class_name="has_ppi_ia",
                        from_rid=from_rid,
                        to_rid=to_rid,
                        value_dict=value_dict,
                    )

                    updated += 1

        return updated



"""BioGRID API methods."""
import json
from collections import namedtuple
from math import ceil
from typing import Dict, List, Optional

from flask import request
from sqlalchemy import and_, or_

from ebel import Bel
from ebel.manager.orientdb.biodbs.biogrid import MODIFICATIONS
from ebel.manager.rdbms.models.biogrid import (Biogrid, ExperimentalSystem,
                                               Interactor, Modification,
                                               Publication, Source, Taxonomy)
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import _get_data

SQL_SELECT_HAS_PPI_BG = """select
    @rid.asString() as rid,
    in.@rid.asString() as subject_rid,
    in.name as subject_name,
    in.namespace as subject_namespace,
    in.label as subject_label,
    in.uniprot as subject_uniprot,
    in.species as subject_taxonomy_id,
    modification,
    pmids,
    biogrid_ids,
    out.@rid.asString() as object_rid,
    out.name as object_name,
    out.namespace as object_namespace,
    out.label as object_label,
    out.uniprot as object_uniprot,
    out.species as object_taxonomy_id
from has_ppi_bg where """


def get_has_ppi_bg_by_symbol_taxid():
    """Get has_ppi_bg edge by POST request with parameter `symbol`."""
    symbol = request.args.get("symbol")
    if not symbol:
        return {"error": "symbol is required"}
    namespace = request.args.get("namespace", "HGNC")
    modification = request.args.get("modification")

    sql = f"""{SQL_SELECT_HAS_PPI_BG}((in.name='{symbol}' and in.namespace='{namespace}')
            or (out.name='{symbol}' and out.namespace='{namespace}'))"""
    if modification and modification in MODIFICATIONS:
        sql += f" and modification = '{modification}'"
    return Bel().query_get_dict(sql)


def get_has_ppi_bg_by_uniprot() -> List[dict]:
    """Get has_ppi_bg edge by POST request with parameter `uniprot`."""
    uniprot = request.args.get("uniprot")
    if not uniprot:
        return [{"error": "uniprot is required"}]
    modification = request.args.get("modification")
    sql = f"""{SQL_SELECT_HAS_PPI_BG}( in.uniprot='{uniprot}' or out.uniprot='{uniprot}')"""
    if modification and modification in MODIFICATIONS:
        sql += f" and modification = '{modification}'"
    return Bel().query_get_dict(sql)


def get_biogrid_by_biogrid_id() -> dict:
    """Get BioGrid entry by biogrid_id.

    Parameters:
        biogrid_id (int): BioGrid identifier

    Returns:
        dict: BioGrid entry.
    """
    biogrid_id: int = request.args.get("biogrid_id")
    biogrid_entry = RDBMS.get_session().query(Biogrid).filter_by(biogrid_id=biogrid_id).first()
    return biogrid_entry.as_dict()


def get_experimental_systems() -> Dict[str, int]:
    """Get all available experimental systems.

    Returns:
        Dict[str, int]: schema: Dict[experimental_system, table_id]
    """
    es = ExperimentalSystem.experimental_system
    rs = RDBMS.get_session().query(es, ExperimentalSystem.id).order_by(es.asc()).all()
    return {x.experimental_system: x.id for x in rs}


def get_sources() -> Dict[str, int]:
    """Get all available sources.

    Returns:
        Dict[str, int]: schema: Dict[source, table_id]
    """
    rs = RDBMS.get_session().query(Source.source, Source.id).order_by(Source.source.asc()).all()
    return {x.source: x.id for x in rs}


def get_taxonomies() -> Dict[str, int]:
    """Get all available taxonomies.

    Returns:
        Dict[str, int]: schema: Dict[organism_name, taxonomy_id]
    """
    org = Taxonomy.organism_name
    rs = RDBMS.get_session().query(org, Taxonomy.taxonomy_id).order_by(org.asc()).all()
    return {x.organism_name: x.taxonomy_id for x in rs}


def get_modifications() -> Dict[str, int]:
    """Get all available modifications.

    Returns:
        Dict[str, int]: schema: Dict[modification, table_id]
    """
    mod = Modification.modification
    rs = RDBMS.get_session().query(Modification.id, mod).order_by(mod.asc()).all()
    return {x.modification: x.id for x in rs}


def get_biogrid_by_pmid() -> List[dict]:
    """List of BioGrid entries by PubMed identifier.

    Parameters:
        pmid (int): PubMed identifier

    Returns:
        List[dict]: List[BioGrid entry]
    """
    pmid = request.args.get("pmid")
    publication_id = (
        RDBMS.get_session().query(Publication.id).filter_by(source="PUBMED", source_identifier=pmid).first()[0]
    )
    biogrid_entry = RDBMS.get_session().query(Biogrid).filter_by(publication_id=publication_id).all()
    return [x.as_dict() for x in biogrid_entry]


def get_biogrid() -> dict:
    """List BioGrid entries by GET request.

    Parameters:
    * id_type_a (str): Type of Interactor A
    * interactor_a (str): Interactor B identifier
    * taxonomy_id_a (int): NCBI taxonomy ID of interactor B
    * id_type_b (str): Type of Interactor B
    * interactor_b (str): Interactor B identifier
    * taxonomy_id_b (int): NCBI taxonomy ID of interactor B
    * experimental_system (str): Experimental system used to identifiy the interaction
    * modification (str): Type of modification
    * source (str): Reference database
    * page_size (int): Number of entries per page
    * page (int): Number of page

    Returns:
        List[dict]: List[BioGrid entry]
    """
    # cancel if neither interactor_a or interactor_b are submitted
    if not (bool(request.args.get("interactor_a")) or bool(request.args.get("interactor_b"))):
        return {"error": "At least interactor_a or interactor_b"}

    Req = namedtuple(
        "RequestObject",
        [
            "id_type_a",
            "interactor_a",
            "taxonomy_id_a",
            "id_type_b",
            "interactor_b",
            "taxonomy_id_b",
            "experimental_system",
            "modification",
            "source",
            "interaction_directed",
            "page_size",
            "page",
        ],
    )
    q = {
        "id_type_a": "symbol",
        "interactor_a": None,
        "taxonomy_id_a": 9606,
        "id_type_b": "symbol",
        "interactor_b": None,
        "taxonomy_id_b": 9606,
        "experimental_system": None,
        "modification": None,
        "source": None,
        "interaction_directed": False,
        "page_size": 10,
        "page": 1,
    }
    for key, value in request.args.items():
        if key in q:
            q[key] = value
    req = Req(**q)

    if req.interactor_a:
        a = {req.id_type_a: req.interactor_a, "taxonomy_id": req.taxonomy_id_a}
        a = {k: v for k, v in a.items() if v}
        biogrid_a_ids = [x[0] for x in RDBMS.get_session().query(Interactor.biogrid_id).filter_by(**a)]

    if req.interactor_b:
        b = {req.id_type_b: req.interactor_b, "taxonomy_id": req.taxonomy_id_b}
        b = {k: v for k, v in b.items() if v}
        biogrid_b_ids = [x[0] for x in RDBMS.get_session().query(Interactor.biogrid_id).filter_by(**b)]

    if req.interactor_a and req.interactor_b:
        interaction_forward = and_(
            Biogrid.biogrid_a_id.in_(biogrid_a_ids),
            Biogrid.biogrid_b_id.in_(biogrid_b_ids),
        )
        if req.interaction_directed:
            query_filter = interaction_forward
        else:
            interaction_backward = and_(
                Biogrid.biogrid_a_id.in_(biogrid_b_ids),
                Biogrid.biogrid_b_id.in_(biogrid_a_ids),
            )
            query_filter = or_(interaction_forward, interaction_backward)
    else:
        if req.interactor_a:
            if req.interaction_directed is True or req.interaction_directed == "true":
                query_filter = Biogrid.biogrid_a_id.in_(biogrid_a_ids)
            else:
                query_filter = or_(
                    Biogrid.biogrid_a_id.in_(biogrid_a_ids),
                    Biogrid.biogrid_b_id.in_(biogrid_a_ids),
                )
        if req.interactor_b:
            if req.interaction_directed is True or req.interaction_directed == "true":
                query_filter = Biogrid.biogrid_b_id.in_(biogrid_b_ids)
            else:
                query_filter = or_(
                    Biogrid.biogrid_a_id.in_(biogrid_b_ids),
                    Biogrid.biogrid_b_id.in_(biogrid_b_ids),
                )

    query = RDBMS.get_session().query(Biogrid).filter(query_filter)

    if req.experimental_system:
        experimental_system_id = (
            RDBMS.get_session()
            .query(ExperimentalSystem.id)
            .filter_by(experimental_system=req.experimental_system)
            .first()[0]
        )
        query = query.filter_by(experimental_system_id=experimental_system_id)

    if req.modification:
        modification_id = (
            RDBMS.get_session().query(Modification.id).filter_by(modification=req.modification).first()[0]
        )
        query = query.filter_by(modification_id=modification_id)

    if req.source:
        source_id = RDBMS.get_session().query(Source.id).filter_by(source=req.source).first()[0]
        query = query.filter_by(source_id=source_id)

    number_of_results = query.count()

    limit = int(req.page_size)
    page = int(req.page)
    offset = (page - 1) * limit
    offset = offset if offset <= 100 else 100
    query = query.limit(limit).offset(offset)
    pages = ceil(number_of_results / limit)

    return {
        "page": page,
        "page_size": limit,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": [x.as_dict() for x in query.all()],
    }


def get_biogrid_using_post():
    """Get BioGRID entry via POSt method."""
    return _get_data(Biogrid)


def get_interactor_by_symbol():
    """Get interactor by gene symbol."""
    ra = request.args

    taxonomy_id = int(ra.get("taxonomy_id", 9606))
    search_term = ra.get("symbol", "")

    query = (
        RDBMS.get_session()
        .query(Interactor)
        .filter(
            Interactor.taxonomy_id == taxonomy_id,
            Interactor.symbol.like(f"{search_term}%"),
        )
    )

    number_of_results = query.count()
    limit = int(ra.get("page_size", 1))
    page = int(ra.get("page", 10))
    offset = (page - 1) * limit
    offset = offset if offset <= 100 else 100
    query = query.limit(limit).offset(offset)
    pages = ceil(number_of_results / limit)

    return {
        "page": page,
        "page_size": limit,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": [x.as_dict() for x in query.all()],
    }



"""CHEBI API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, RESULTS


class TestChebi:
    def test_get_compound_by_name(self, client):
        response = client.get("api/v1/chebi/compound/by_name?name=donepezil", content_type="application/json")
        output = format_response_data(response)
        expected_results = {  # only check some of the generated information
            "source": "ChEBI",
            "status": "C",
            "chebi_accession": "CHEBI:53289",
            "name": "donepezil",
        }
        assert isinstance(output, list)
        hit = output[0]
        assert isinstance(hit, dict)
        for key, value in expected_results.items():
            assert key in hit
            assert hit[key] == value

    def test_get_compound_name_by_name_starts_with(self, client):
        response = client.get(
            "api/v1/chebi/compound_name/by_name_starts_with?name=donep", content_type="application/json"
        )
        output = format_response_data(response)
        expected_results = {"donepezil": 53289, "donepezil (1+)": 145498, "donepezil hydrochloride": 4696}
        assert isinstance(output, dict)
        assert expected_results == output

    def test_get_compound_by_id(self, client):
        response = client.get("api/v1/chebi/compound/by_id?id=53289", content_type="application/json")
        output = format_response_data(response)
        expected_results = {  # only check some of the generated information
            "source": "ChEBI",
            "status": "C",
            "chebi_accession": "CHEBI:53289",
            "name": "donepezil",
        }
        assert isinstance(output, dict)
        for key, value in expected_results.items():
            assert key in output
            assert output[key] == value

    def test_get_compound_by_other_db_accession(self, client):
        response = client.get(
            "api/v1/chebi/compound/by_other_db_accession?db_name=DrugBank%20accession&accession_number=DB00843",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {  # only check some of the generated information
            "status": "C",
            "chebi_accession": "CHEBI:4696",
            "name": "donepezil hydrochloride",
        }
        assert isinstance(output, list)
        hit = output[0]
        assert isinstance(hit, dict)
        for key, value in expected_results.items():
            assert key in hit
            assert hit[key] == value

    def test_get_compound_reference(self, client):
        response = client.get(
            "api/v1/chebi/compound_reference?reference_id=US2006025345&reference_db_name=Patent&reference_name=Substituted%20ethane-1%2C2-diamines%20for%20the%20treatment%20of%20Alzheimer%27s%20disease&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "compound_id": 17596,
            "location_in_ref": None,
            "reference_db_name": "Patent",
            "reference_id": "US2006025345",
            "reference_name": "Substituted ethane-1,2-diamines for the treatment of Alzheimer's disease",
        }
        assert output[NUM_RESULTS] > 20
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_relation(self, client):
        response = client.get("api/v1/chebi/relation?final_id=53289", content_type="application/json")
        output = format_response_data(response)
        assert isinstance(output, list)
        assert len(output) > 0
        first_entry = output[0]
        expected_cols = ["final_id", "init_id", "status", "type"]
        assert all([col in first_entry for col in expected_cols])

    def test_get_bel_chebi_ids(self, client):
        response = client.get(
            "api/v1/chebi/ebel/nodes?name=ATP&namespace=CHEBI&chebi=15422&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        assert isinstance(output, dict)
        results = output[RESULTS]
        assert isinstance(results, list)
        assert len(output) >= 5
        hit = results[0]
        assert all([col in hit for col in ("bel", "chebi", "name", "namespace", "rid")])
        assert hit["name"] == "ATP"
        assert hit["chebi"] == 15422



"""Chebi."""
import logging
import re
from typing import Dict

import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel import tools
from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.orientdb.constants import CHEBI
from ebel.manager.rdbms.models import chebi

logger = logging.getLogger(__name__)


class Chebi(odb_meta.Graph):
    """Chebi."""

    def __init__(self, client: OrientDB = None):
        """Init CHEBI."""
        self.client = client
        self.biodb_name = CHEBI
        self.urls = {
            chebi.Compound.__tablename__: urls.CHEBI_COMPOUND,
            chebi.Relation.__tablename__: urls.CHEBI_RELATION,
            chebi.Inchi.__tablename__: urls.CHEBI_INCHI,
            chebi.ChemicalData.__tablename__: urls.CHEBI_CHEMICALDATA,
            chebi.Comment.__tablename__: urls.CHEBI_COMMENT,
            chebi.DatabaseAccession.__tablename__: urls.CHEBI_DATABASEACCESSION,
            chebi.Name.__tablename__: urls.CHEBI_NAME,
            chebi.Reference.__tablename__: urls.CHEBI_REFERENCE,
            chebi.Structure.__tablename__: urls.CHEBI_STRUCTURE,
        }

        super().__init__(urls=self.urls, biodb_name=self.biodb_name, tables_base=chebi.Base)

    def __len__(self) -> int:
        """Get number of edges in OrientDB."""
        sql = "Select count(*) from bel where namespace='CHEBI'"
        return self.execute(sql)[0].oRecordData["count"]

    def __contains__(self, name: str) -> bool:
        """Checks if RS number (without prefix RS) exists in BEL graph."""
        sql = f'Select 1 from bel where namespace="CHEBI" and name = "{name}"'
        return bool(self.execute(sql))

    def insert_data(self) -> Dict[str, int]:
        """Insert data in generic OrientDB class."""
        logger.info(f"Insert data in {self.biodb_name}")

        inserted = {}

        file_path_compound = tools.get_file_path(urls.CHEBI_COMPOUND, self.biodb_name)
        df_compounds = pd.read_csv(file_path_compound, sep="\t", low_memory=False, on_bad_lines="skip")
        df_compound_ids = df_compounds[["ID"]]
        del df_compounds

        for table_name, url in tqdm(self.urls.items(), desc=f"Import {self.biodb_name.upper()}"):
            file_path = tools.get_file_path(url, self.biodb_name)
            seperator = "\t" if re.search(r".*\.tsv(\.gz)?$", file_path) else ","
            encoding = "ISO-8859-1" if table_name == "chebi_reference" else None

            dfs = pd.read_csv(
                file_path,
                sep=seperator,
                encoding=encoding,
                low_memory=False,
                on_bad_lines="skip",
                chunksize=100000,
            )
            inserted[table_name] = 0
            for df in dfs:
                df.columns = df.columns.str.lower()

                if table_name == chebi.Inchi.__tablename__:
                    # with_index = True
                    df.index += 1
                    df.index.rename("id", inplace=True)
                    df.rename(columns={"chebi_id": "compound_id"}, inplace=True)

                if "compound_id" in df.columns:
                    df = (
                        df_compound_ids.rename(columns={"ID": "compound_id"})
                        .set_index("compound_id")
                        .join(df.set_index("compound_id"), how="inner")
                        .reset_index()
                    )

                if "init_id" in df.columns:
                    df = (
                        df_compound_ids.rename(columns={"ID": "init_id"})
                        .set_index("init_id")
                        .join(df.set_index("init_id"), how="inner")
                        .reset_index()
                    )

                if "final_id" in df.columns:
                    df = (
                        df_compound_ids.rename(columns={"ID": "final_id"})
                        .set_index("final_id")
                        .join(df.set_index("final_id"), how="inner")
                        .reset_index()
                    )

                df.to_sql(table_name, self.engine, index=False, if_exists="append")

                inserted[table_name] += df.shape[0]
        self.session.commit()
        return inserted

    def update_bel(self) -> int:
        """Update the BEL edges with CHEBI metadata."""
        updated = 0
        sql = "Update {rid} set chebi = {chebi_id}"
        chebi_nodes = self.query_class(
            class_name="bio_object",
            columns=["name", "@class"],
            namespace="CHEBI",
            pure=True,
        )

        for chebi_node in tqdm(chebi_nodes, desc="Update ChEBI identifier in BEL"):
            chebi_compound = (
                self.session.query(chebi.Compound.id).filter(chebi.Compound.name == chebi_node["name"]).first()
            )

            if chebi_compound:
                updated += self.execute(sql.format(rid=chebi_node["rid"], chebi_id=chebi_compound[0]))[0]

        return updated

    def update_interactions(self) -> int:
        """Abstract method."""
        return 0



"""DrugBank API methods."""
from flask import request
from sqlalchemy.orm.attributes import InstrumentedAttribute

from ebel.manager.rdbms.models import drugbank
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_data, _get_paginated_ebel_query_result,
                                  _get_paginated_query_result)


def get_by_id():
    """Get DrugBank entry by ID."""
    drugbank_id = request.args.get("drugbank_id")
    if drugbank_id:
        query = RDBMS.get_session().query(drugbank.Drugbank).filter_by(drugbank_id=drugbank_id.strip())
        return query.first().as_dict()


def get_drugbank():
    """Get generic DrugBank entry."""
    return _get_data(drugbank.Drugbank)


def get_interaction():
    """Get DrugBank drug interactions."""
    return _get_data(drugbank.DrugInteraction)


def _get_model(model):
    """Get tables connected to Drugbank - paginated."""
    drugbank_id = request.args.get("drugbank_id")
    columns_like = {
        col_obj.like(request.args[col_name])
        for col_name, col_obj in model.__dict__.items()
        if isinstance(col_obj, InstrumentedAttribute) and col_name in request.args and col_name != "drugbank_id"
    }
    query = RDBMS.get_session().query(model).filter(*columns_like)
    if drugbank_id:
        query = query.join(drugbank.Drugbank).filter_by(drugbank_id=drugbank_id)
    return _get_paginated_query_result(query, print_sql=True)


def get_pathway():
    """Get DrugBank involved pathways."""
    return _get_model(drugbank.Pathway)


def get_status():
    """Get DrugBank statuses."""
    return _get_model(drugbank.Status)


def get_patent():
    """Get DrugBank patents."""
    return _get_model(drugbank.Patent)


def get_external_identifier():
    """Get DrugBank external identifiers."""
    return _get_model(drugbank.ExternalIdentifier)


def get_reference():
    """Get DrugBank references."""
    return _get_model(drugbank.Reference)


def get_target():
    """Get DrugBank targets."""
    return _get_model(drugbank.Target)


def get_product_name():
    """Get DrugBank product names."""
    return _get_model(drugbank.ProductName)


def get_synonym():
    """Get DrugBank synonyms."""
    return _get_model(drugbank.Synonym)


def get_has_drug_target_db():
    """Get DrugBank related eBEL relations."""
    conf = {
        "target_rid": "in.@rid.asString()",
        "target_name": "in.name",
        "target_namespace": "in.namespace",
        "target_uniprot_accession": "in.uniprot",
        "drugbank_id": "out.drugbank_id",
        "drug_name": "out.label",
        "drug_rid": "out.@rid.asString()",
        "edge_rid": "@rid.asString()",
        "action": "action",
        "known_action": "known_action",
    }
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_drug_target_db"
    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        sql += " WHERE " + " AND ".join([f'{conf[k]} = "{v}"' for k, v in paras.items()])
    return _get_paginated_ebel_query_result(sql)



"""BioGRID RDBMS model definition."""
from sqlalchemy import Column, Float, ForeignKey, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Biogrid(Base):
    """Class definition for the biogrid table."""

    __tablename__ = "biogrid"
    id = Column(Integer, primary_key=True)

    biogrid_a_id = Column(Integer, ForeignKey("biogrid_interactor.biogrid_id"))
    biogrid_a = relationship("Interactor", foreign_keys=[biogrid_a_id])
    biogrid_b_id = Column(Integer, ForeignKey("biogrid_interactor.biogrid_id"))
    biogrid_b = relationship("Interactor", foreign_keys=[biogrid_b_id])
    biogrid_id = Column(Integer, nullable=True)
    experimental_system_id = Column(Integer, ForeignKey("biogrid_experimental_system.id"))
    experimental_system = relationship("ExperimentalSystem", foreign_keys=[experimental_system_id])
    throughput_id = Column(Integer, ForeignKey("biogrid_throughput.id"))
    throughput = relationship("Throughput", foreign_keys=[throughput_id])
    score = Column(Float, nullable=True)
    modification_id = Column(Integer, ForeignKey("biogrid_modification.id"))
    modification = relationship("Modification", foreign_keys=[modification_id])
    qualifications = Column(String(255), nullable=True)
    source_id = Column(Integer, ForeignKey("biogrid_source.id"))
    source = relationship("Source", foreign_keys=[source_id])
    publication_id = Column(Integer, ForeignKey("biogrid_publication.id"))
    publication = relationship("Publication", foreign_keys=[publication_id])
    qualification = Column(Text, nullable=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "biogrid_a": self.biogrid_a.as_dict(),
            "biogrid_b": self.biogrid_b.as_dict(),
            "experimental_system": self.experimental_system.as_dict() if self.experimental_system else None,
            "throughput": self.throughput.as_dict() if self.throughput else None,
            "biogrid_id": self.biogrid_id,
            "score": self.score if self.score else None,
            "modification": self.modification.as_dict() if self.modification else None,
            "source": self.source.source,
            "publication": self.publication.as_dict(),
            "qualification": self.qualification if self.qualification else None,
        }


class Publication(Base):
    """Class definition for the biogrid_publication table."""

    __tablename__ = "biogrid_publication"
    id = Column(Integer, primary_key=True)
    author_name = Column(String(255), nullable=True)
    publication_year = Column(Integer, nullable=True)
    source = Column(String(255), nullable=True)
    source_identifier = Column(String(255), nullable=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])


class Throughput(Base):
    """Class definition for the biogrid_throughput table."""

    __tablename__ = "biogrid_throughput"
    id = Column(Integer, primary_key=True)
    throughput = Column(String(255))
    frequency = Column(Integer)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])


class Taxonomy(Base):
    """Class definition for the biogrid_taxonomy table."""

    __tablename__ = "biogrid_taxonomy"
    taxonomy_id = Column(Integer, primary_key=True)  # == NCBI Taxonomy ID
    organism_name = Column(String(1000))

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"taxonomy_id": self.taxonomy_id, "organism_name": self.organism_name}


class ExperimentalSystem(Base):
    """Class definition for the biogrid_experimental_system table."""

    __tablename__ = "biogrid_experimental_system"
    id = Column(Integer, primary_key=True)
    experimental_system = Column(String(255), nullable=True)
    experimental_system_type = Column(String(255), nullable=True)
    frequency = Column(Integer)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])


class Interactor(Base):
    """Class definition for the biogrid_interactor table."""

    __tablename__ = "biogrid_interactor"
    biogrid_id = Column(Integer, primary_key=True)

    entrez = Column(Integer, nullable=True, index=True)
    systematic_name = Column(String(255), nullable=True, index=True)
    symbol = Column(String(255), nullable=True, index=True)
    taxonomy_id = Column(Integer, ForeignKey("biogrid_taxonomy.taxonomy_id"))
    taxonomy = relationship("Taxonomy", foreign_keys=[taxonomy_id])
    uniprot = Column(String(255), nullable=True, index=True)
    trembl = Column(String(1000), nullable=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "entrez": self.entrez,
            "systematic_name": self.systematic_name,
            "symbol": self.symbol,
            "uniprot": self.uniprot,
            "trembl": self.trembl,
            "taxonomy": self.taxonomy.as_dict(),
        }


class Source(Base):
    """Class definition for the biogrid_source table."""

    __tablename__ = "biogrid_source"
    id = Column(Integer, primary_key=True)
    source = Column(String(255), nullable=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])


class Modification(Base):
    """Class definition for the biogrid_modification table."""

    __tablename__ = "biogrid_modification"
    id = Column(Integer, primary_key=True)
    modification = Column(String(255), nullable=True)
    frequency = Column(Integer)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self, exclude=["id"])



"""Clinical Trials."""
import logging
from collections import namedtuple
from typing import Dict, Optional
from zipfile import ZipFile

import pandas as pd
from lxml import etree
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel import tools
from ebel.constants import RID
from ebel.manager.orientdb import odb_meta, urls
from ebel.manager.orientdb.constants import CLINICAL_TRIALS
from ebel.manager.rdbms.models import clinical_trials_gov

Intervention = namedtuple("Intervention", ["intervention_type", "intervention_name"])

logger = logging.getLogger(__name__)


class ClinicalTrials(odb_meta.Graph):
    """ClinicalTrials.gov."""

    def __init__(self, client: OrientDB = None, condition_keyword="Alzheimer"):
        """Init ClinicalTrials."""
        self.condition_keyword = condition_keyword
        self.client = client
        self.biodb_name = CLINICAL_TRIALS
        self.url = urls.CLINICAL_TRIALS_GOV
        self.urls = {self.biodb_name: self.url}
        self.file_path = tools.get_file_path(urls.CLINICAL_TRIALS_GOV, self.biodb_name)
        super().__init__(
            tables_base=clinical_trials_gov.Base,
            # nodes=odb_structure.clinical_trials_gov_nodes,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def add_link_to_drugbank(self, data_dict: dict, trial_rid: str):
        """Create LINKSET in drugbank table for associated clinical trials."""
        # Can't check synonyms untils OrientDB 3.0, need to be able to index on collections
        # update_sql = 'UPDATE drugbank ADD clinical_trials = {} WHERE name = "{}" OR "{}" in synonyms'
        # TODO index drugbank.synonyms

        results = self.query_class(class_name="drugbank", limit=1)  # Check if drugbank is there

        if results:
            update_sql = 'UPDATE drugbank ADD clinical_trials = {} WHERE name = "{}"'

            if "drugs_in_trial" in data_dict.keys():
                for drug in data_dict["drugs_in_trial"]:
                    drug = drug.replace('"', "'")  # Have to scrub the string
                    # self.execute(update_sql.format(trial_rid, drug, drug))
                    self.execute(update_sql.format(trial_rid, drug))

    def insert_n2m_tables(self):
        """Inserts mesh_terms and keywords."""
        with ZipFile(self.file_path, "r") as zip_file:
            xml_files = [x for x in zip_file.filelist if x.filename.endswith(".xml")]

            d = {
                "keyword": set(),
                "mesh_term": set(),
                "condition": set(),
                "intervention": set(),
            }
            interventions = set()

            for f in tqdm(xml_files, desc="Get and set unique values tables in ClinicalTrials.gov"):
                with zip_file.open(f.filename) as xml_file:
                    xml_content = xml_file.read()

                doc = etree.fromstring(xml_content)

                for child in doc:
                    if child.tag in ["keyword", "condition"]:
                        d[child.tag].add(child.text)

                    elif child.tag == "condition_browse":
                        d["mesh_term"].update(set(child.xpath("./mesh_term//text()")))

                    elif child.tag == "intervention":
                        itype, iname = self.__parse_intervention(child)
                        interventions.add(
                            Intervention(
                                intervention_type=(itype.strip() if itype else itype),
                                intervention_name=(iname.strip() if iname else iname),
                            )
                        )

        df_interventions = pd.DataFrame(interventions)
        df_interventions.index += 1
        df_interventions.index.rename("id", inplace=True)
        df_interventions.to_sql(
            clinical_trials_gov.Intervention.__tablename__,
            self.engine,
            if_exists="append",
        )

        for column_name, data in d.items():
            df = pd.DataFrame(data, columns=[column_name])
            df.index += 1
            df.index.rename("id", inplace=True)
            df.to_sql("clinical_trials_gov_" + column_name, self.engine, if_exists="append")

    @staticmethod
    def __parse_intervention(interventions) -> tuple:
        """Extract the intervention type and intervention name from the xml child."""
        intervention = dict()

        for cchild in interventions:
            intervention[cchild.tag] = cchild.text.strip()

        itype = intervention.get("intervention_type")
        iname = intervention.get("intervention_name")

        return itype, iname

    @staticmethod
    def __get_first(element: list) -> Optional[str]:
        if len(element):
            return str(element[0])

    def insert_mesh_terms(self, df):
        """Insert mesh_terms into database."""
        columns = {
            "id": "clinical_trials_gov_id",
            "mesh_terms": "clinical_trials_gov_mesh_term_id",
        }
        table = clinical_trials_gov.ctg_mesh_term_n2m.name
        df[["mesh_terms", "id"]].explode("mesh_terms").rename(columns=columns).to_sql(
            table, index=False, if_exists="append", con=self.engine
        )

    def insert_keywords(self, df):
        """Insert keywords into database."""
        columns = {
            "id": "clinical_trials_gov_id",
            "keywords": "clinical_trials_gov_keyword_id",
        }
        table = clinical_trials_gov.ctg_keyword_n2m.name
        df[["keywords", "id"]].explode("keywords").rename(columns=columns).to_sql(
            table, index=False, if_exists="append", con=self.engine
        )

    def insert_conditions(self, df):
        """Insert conditions into database."""
        columns = {
            "id": "clinical_trials_gov_id",
            "conditions": "clinical_trials_gov_condition_id",
        }
        table = clinical_trials_gov.ctg_condition_n2m.name
        df[["conditions", "id"]].explode("conditions").rename(columns=columns).to_sql(
            table, index=False, if_exists="append", con=self.engine
        )

    def insert_interventions(self, df):
        """Insert interventions into database."""
        columns = {
            "id": "clinical_trials_gov_id",
            "interventions": "clinical_trials_gov_intervention_id",
        }
        table = clinical_trials_gov.ctg_intervention_n2m.name
        df[["interventions", "id"]].explode("interventions").rename(columns=columns).to_sql(
            table, index=False, if_exists="append", con=self.engine
        )

    def insert_data(self) -> Dict[str, int]:
        """Insert Clinical Trial metadata into database."""
        self.recreate_tables()
        with ZipFile(self.file_path, "r") as zip_file:
            xml_files = [x for x in zip_file.filelist if x.filename.endswith(".xml")]

            self.insert_n2m_tables()
            conditions = {x.condition: x.id for x in self.session.query(clinical_trials_gov.Condition).all()}
            keywords = {x.keyword: x.id for x in self.session.query(clinical_trials_gov.Keyword).all()}
            mesh_terms = {x.mesh_term: x.id for x in self.session.query(clinical_trials_gov.MeshTerm).all()}
            interventions = {
                (x.intervention_type, x.intervention_name): x.id
                for x in self.session.query(clinical_trials_gov.Intervention).all()
            }

            trials = []
            index = 0

            for index, f in tqdm(
                enumerate(xml_files, 1),
                desc=f"Import {self.biodb_name.upper()}",
                total=len(xml_files),
            ):
                with zip_file.open(f.filename) as xml_file:
                    xml_content = xml_file.read()

                doc = etree.fromstring(xml_content)
                data_dict = self.get_data_as_dict(doc, conditions, keywords, mesh_terms, interventions)
                data_dict["id"] = index
                trials.append(data_dict)

                if index % 10000 == 0:
                    self.insert_trials(trials)
                    trials = []

            self.insert_trials(trials)

        return {self.biodb_name: index}

    def insert_trials(self, trials):
        """Insert select Clinical Trial entries into database."""
        cols_multi = [
            "primary_outcomes",
            "secondary_outcomes",
            "interventions",
            "mesh_terms",
            "keywords",
            "conditions",
        ]
        df = pd.DataFrame(trials)
        cols = list(set(df.columns) - set(cols_multi))
        df[cols].to_sql(
            clinical_trials_gov.ClinicalTrialGov.__tablename__,
            self.engine,
            if_exists="append",
            index=False,
        )
        self.insert_mesh_terms(df)
        self.insert_keywords(df)
        self.insert_conditions(df)
        self.insert_interventions(df)

    def get_data_as_dict(self, doc, conditions, keywords, mesh_terms, interventions):
        """Get metadata as dict based on passed parameters."""
        d = {
            "primary_outcomes": [],
            "secondary_outcomes": [],
            "keywords": [],
            "conditions": [],
            "interventions": [],
            "mesh_terms": [],
        }
        for child in doc:
            if child.tag == "id_info":
                d["nct_id"] = self.__get_first(child.xpath("./nct_id[1]/text()"))
                d["org_study_id"] = self.__get_first(child.xpath("./org_study_id[1]/text()"))

            elif child.tag in [
                "brief_title",
                "official_title",
                "overall_status",
                "start_date",
                "completion_date",
                "phase",
                "study_type",
            ]:
                d[child.tag] = child.text

            elif child.tag in ["brief_summary", "detailed_description"]:
                d[child.tag] = self.__get_first(child.xpath("./textblock[1]/text()"))

            elif child.tag == "oversight_info":
                d["is_fda_regulated_drug"] = self.__get_first(child.xpath("./is_fda_regulated_drug[1]/text()"))

            elif child.tag == "condition":
                d["conditions"].append(conditions[child.text])

            elif child.tag == "condition_browse":
                d["mesh_terms"] = [mesh_terms[x] for x in child.xpath("./mesh_term//text()")]

            elif child.tag == "study_design_info":
                d["study_design_intervention_model"] = self.__get_first(child.xpath("./intervention_model[1]/text()"))
                d["study_design_primary_purpose"] = self.__get_first(child.xpath("./primary_purpose[1]/text()"))
                d["study_design_masking"] = self.__get_first(child.xpath("./masking[1]/text()"))

            elif child.tag in ["primary_outcome", "secondary_outcome"]:
                outcomes = dict()
                for cchild in child:
                    outcomes[cchild.tag] = cchild.text
                d[child.tag + "s"] = outcomes

            elif child.tag == "keyword":
                d["keywords"].append(keywords[child.text])

            elif child.tag == "intervention":
                itype, iname = self.__parse_intervention(child)
                formatted_intervention = interventions[(itype, iname)]
                d["interventions"].append(formatted_intervention)

            elif child.tag == "patient_data":
                d["patient_data_sharing_ipd"] = self.__get_first(child.xpath("./sharing_ipd[1]/text()"))
                d["patient_data_ipd_description"] = self.__get_first(child.xpath("./ipd_description[1]/text()"))

        return d

    def update_bel(self):
        """Update the BEL relations with links to pathologies."""
        # TODO FIX: Too slow!! Need to be able to index on collections
        # self.update_pathology_links()
        pass

    def update_pathology_links(self) -> int:
        """Update the pathology class to link to associated clinical trials."""
        trial_sql = "SELECT @rid.asString() FROM clinical_trial WHERE '{}' in mesh_conditions"
        update_sql = "UPDATE {} SET clinical_trials = {}"

        paths = self.query_class(
            class_name="pathology",
            columns=["name"],
        )

        updated = 0
        for path in tqdm(paths, desc="Update pathology nodes"):
            path_rid, path_name = path[RID], path["name"]

            trial_results = self.client.query(trial_sql.format(path_name))
            trial_rids = [x.oRecordData[RID] for x in trial_results]
            trial_linkset = "[" + ",".join(trial_rids) + "]"  # This feature is a LINKSET

            self.execute(update_sql.format(path_rid, trial_linkset))
            updated += 1

        return updated

    def update_interactions(self) -> int:
        """Abstract method."""
        pass



"""String constants for the OrientDB submodule."""


# Databases
BIOGRID = "biogrid"
CHEBI = "chebi"
CLINICAL_TRIALS = "clinical_trials"
CLINVAR = "clinvar"
DISGENET = "disgenet"
DRUGBANK = "drugbank"
ENSEMBL = "ensembl"
FLYBASE = "flybase"
GWAS_CATALOG = "gwas_catalog"
HGNC = "hgnc"
EXPRESSION_ATLAS = "expression_atlas"
INTACT = "intact"
IUPHAR = "iuphar"
KEGG = "kegg"
MGI = "mgi"
MIRTARBASE = "mirtarbase"
NCBI = "ncbi"
OFFSIDES = "offsides"
ONSIDES = "onsides"
NSIDES = "nsides"
PATHWAY_COMMONS = "pathway_commons"
PROTEIN_ATLAS = "protein_atlas"
REACTOME = "reactome"
RGD = "rgd"
SIDER = "sider"
STITCH = "stitch"
STRINGDB = "stringdb"
UNIPROT = "uniprot"

# Data Types
NODES = "nodes"
EDGES = "edges"



"""HGNC Human Ortholog RDBMS model definition."""
from sqlalchemy import Column, Integer, String, Text
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class HumanOrtholog(Base):
    """Class definition for the human_ortholog table."""

    __tablename__ = "human_ortholog"

    id = Column(Integer, primary_key=True)

    hgnc_id = Column(String(20), index=True)
    human_entrez_gene = Column(Integer)
    human_ensembl_gene = Column(String(20))
    human_symbol = Column(String(50), index=True)
    ortholog_species = Column(Integer, index=True)
    ortholog_species_entrez_gene = Column(Integer)
    ortholog_species_ensembl_gene = Column(String(50))
    ortholog_species_db_id = Column(String(50))
    ortholog_species_symbol = Column(String(50), index=True)
    support = Column(Text)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)



"""miRTarBase API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestMirTarBase:
    def test_get_mirtarbase(self, client):
        response = client.get(
            "api/v1/mirtarbase?mi_rtar_base_id=MIRT019168&mi_rna=hsa-miR-335-5p&species_mi_rna=Homo%20sapiens&target_gene=CD33&target_gene_entrez_id=945&species_target_gene=Homo%20sapiens&experiments=Microarray&support_type=Functional%20MTI%20%28Weak%29&references_pmid=18185580&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "experiments": "Microarray",
            "id": 15465,
            "mi_rna": "hsa-miR-335-5p",
            "mi_rtar_base_id": "MIRT019168",
            "references_pmid": 18185580,
            "species_mi_rna": "Homo sapiens",
            "species_target_gene": "Homo sapiens",
            "support_type": "Functional MTI (Weak)",
            "target_gene": "CD33",
            "target_gene_entrez_id": 945,
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""NSIDES RDBMS model definition."""

from sqlalchemy import Column, Float, Index, Integer, String
from sqlalchemy.ext.declarative import declarative_base

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Nsides(Base):
    """Class definition for the nSIDES table."""

    __tablename__ = "nsides"
    __table_args__ = (
        Index(
            "idx_nsides_multi",
            "condition_meddra_id",
            "condition_concept_name",
            "prr",
            "mean_reporting_frequency",
        ),
    )
    id = Column(Integer, primary_key=True)

    drug_rxnorn_id = Column(String(20), index=True)  # This has to be a String because of mapping to drugbank ids
    drug_concept_name = Column(String(255), index=True)

    source = Column(String(10), index=True)

    condition_meddra_id = Column(Integer)
    condition_concept_name = Column(String(255), index=True)

    # OFFSIDES specific
    a = Column(Integer)
    b = Column(Integer)
    c = Column(Integer)
    d = Column(Integer)
    prr = Column(Float)
    prr_error = Column(Float)
    mean_reporting_frequency = Column(Float, index=True)

    def as_dict(self):
        """Convert object values to dictionary."""
        return object_as_dict(self)



"""UniProt RDBMS model definition."""
from collections import defaultdict

from sqlalchemy import Column, ForeignKey, Integer, String, Table, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()

uniprot__uniprot_keyword = Table(
    "uniprot__uniprot_keyword",
    Base.metadata,
    Column("uniprot_id", Integer, ForeignKey("uniprot.id")),
    Column("uniprot_keyword_id", Integer, ForeignKey("uniprot_keyword.keywordid")),
)

uniprot__uniprot_host = Table(
    "uniprot__uniprot_host",
    Base.metadata,
    Column("uniprot_id", Integer, ForeignKey("uniprot.id")),
    Column("uniprot_organism_id", Integer, ForeignKey("uniprot_organism.taxid")),
)

uniprot__uniprot_xref = Table(
    "uniprot__uniprot_xref",
    Base.metadata,
    Column("uniprot_id", Integer, ForeignKey("uniprot.id")),
    Column("uniprot_xref_id", Integer, ForeignKey("uniprot_xref.id")),
)

uniprot__uniprot_subcellular_location = Table(
    "uniprot__uniprot_subcellular_location",
    Base.metadata,
    Column("uniprot_id", Integer, ForeignKey("uniprot.id")),
    Column(
        "uniprot_subcellular_location_id",
        Integer,
        ForeignKey("uniprot_subcellular_location.id"),
    ),
)


class Uniprot(Base):
    """Class definition for the UniProt table."""

    __tablename__ = "uniprot"

    id = Column(Integer, primary_key=True)

    accession = Column(String(20), unique=True)
    name = Column(String(100), nullable=False, unique=True)
    recommended_name = Column(String(255), nullable=True)

    taxid = Column(Integer, ForeignKey("uniprot_organism.taxid"), nullable=False, index=True)
    organism = relationship("Organism")

    function_id = Column(Integer, ForeignKey("uniprot_function.id"), nullable=True)
    function = relationship("Function")

    gene_names = relationship("Gene", back_populates="uniprot")

    gene_symbol = relationship("GeneSymbol", uselist=False, back_populates="uniprot")

    keywords = relationship("Keyword", secondary=uniprot__uniprot_keyword, back_populates="uniprots")

    hosts = relationship("Organism", secondary=uniprot__uniprot_host, back_populates="uniprots")

    xrefs = relationship("Xref", secondary=uniprot__uniprot_xref, back_populates="uniprots")

    subcellular_locations = relationship(
        "SubcellularLocation",
        secondary=uniprot__uniprot_subcellular_location,
        back_populates="uniprots",
    )

    def __repr__(self):
        return self.name

    def as_dict(self):
        """Convert object values to dictionary."""
        xrefs_grouped = defaultdict(list)
        for xref in self.xrefs:
            xrefs_grouped[xref.db].append(xref.identifier)
        xrefs_grouped = {k: sorted(v) for k, v in xrefs_grouped.items()}

        return {
            "name": self.name,
            "accession": self.accession,
            "recommended_name": self.recommended_name,
            "taxid": self.taxid,
            "function_description": self.function.description if self.function else self.function,
            "gene_names": [x.name for x in self.gene_names],
            "gene_symbol": self.gene_symbol.symbol if self.gene_symbol else self.gene_symbol,
            "keywords": [{"keyword": x.keyword_name, "id": x.keywordid} for x in self.keywords],
            "hosts": [{"name": x.scientific_name, "taxid": x.taxid} for x in self.hosts],
            "xrefs": xrefs_grouped,
            "subcellular_locations": [x.name for x in self.subcellular_locations],
            "organism": self.organism.scientific_name,
        }


class GeneSymbol(Base):
    """Class definition for the uniprot_gene_symbol table."""

    __tablename__ = "uniprot_gene_symbol"
    id = Column(Integer, primary_key=True)
    symbol = Column(String(100), nullable=False, index=True)
    uniprot_id = Column(Integer, ForeignKey("uniprot.id"))
    uniprot = relationship("Uniprot", back_populates="gene_symbol")

    def __repr__(self):
        """Define repr."""
        return self.symbol


class Gene(Base):
    """Class definition for the uniprot_gene table."""

    __tablename__ = "uniprot_gene"
    id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False, index=True)
    uniprot_id = Column(Integer, ForeignKey("uniprot.id"))
    uniprot = relationship("Uniprot", back_populates="gene_names")


class Keyword(Base):
    """Class definition for the uniprot_keyword table."""

    __tablename__ = "uniprot_keyword"

    keywordid = Column(Integer, primary_key=True)
    keyword_name = Column(String(100), index=True)

    uniprots = relationship("Uniprot", secondary=uniprot__uniprot_keyword, back_populates="keywords")

    def __repr__(self):
        """Define repr."""
        return f"{self.keyword_name}[{self.keywordid}]"


class Organism(Base):
    """Class definition for the uniprot_organism table."""

    __tablename__ = "uniprot_organism"

    taxid = Column(Integer, primary_key=True)
    scientific_name = Column(String(255))  # TODO:Check if index=True with  is possible

    uniprots = relationship("Uniprot", secondary=uniprot__uniprot_host, back_populates="hosts")


class SubcellularLocation(Base):
    """Class definition for the uniprot_subcellular_location table."""

    __tablename__ = "uniprot_subcellular_location"

    id = Column(Integer, primary_key=True)

    name = Column(String(100), index=True)

    uniprots = relationship(
        "Uniprot",
        secondary=uniprot__uniprot_subcellular_location,
        back_populates="subcellular_locations",
    )


class Xref(Base):
    """Class definition for the uniprot_xref table."""

    __tablename__ = "uniprot_xref"

    id = Column(Integer, primary_key=True)

    db = Column(String(50), index=True)
    identifier = Column(String(100), index=True)

    uniprots = relationship("Uniprot", secondary=uniprot__uniprot_xref, back_populates="xrefs")


class Function(Base):
    """Class definition for the uniprot_function table."""

    __tablename__ = "uniprot_function"

    id = Column(Integer, primary_key=True)

    description = Column(Text)

    uniprots = relationship("Uniprot", back_populates="function")



"""StringDB API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestStringdb:
    def test_get_stringdb(self, client):
        response = client.get(
            "api/v1/string?protein1=9606.ENSP00000000412&protein2=9606.ENSP00000000233&symbol1=M6PR&symbol2=ARF5&neighborhood=0&neighborhood_transferred=0&fusion=0&cooccurence=0&homology=0&coexpression=82&coexpression_transferred=61&experiments=0&experiments_transferred=0&database=0&database_transferred=0&textmining=105&textmining_transferred=0&combined_score=161&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)

        expected_results = {
            "coexpression": 82,
            "coexpression_transferred": 61,
            "combined_score": 161,
            "cooccurence": 0,
            "database": 0,
            "database_transferred": 0,
            "experiments": 0,
            "experiments_transferred": 0,
            "fusion": 0,
            "homology": 0,
            "neighborhood": 0,
            "neighborhood_transferred": 0,
            "protein1": "9606.ENSP00000000412",
            "protein2": "9606.ENSP00000000233",
            "symbol1": "M6PR",
            "symbol2": "ARF5",
            "textmining": 105,
            "textmining_transferred": 0,
        }

        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_stringdb_action(self, client):
        response = client.get(
            "api/v1/string/action?item_id_a=9606.ENSP00000216366&item_id_b=9606.ENSP00000000233&symbol1=AP4S1&symbol2=ARF5&mode=binding&score=165&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)

        expected_results = {
            "a_is_acting": False,
            "action": None,
            "is_directional": False,
            "item_id_a": "9606.ENSP00000216366",
            "item_id_b": "9606.ENSP00000000233",
            "mode": "binding",
            "score": 165,
            "symbol1": "AP4S1",
            "symbol2": "ARF5",
        }

        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""Variables used by all tests in the test_manager directory."""
import os
import pathlib

from ebel import Bel
from ebel.config import get_config_value
from ebel.constants import DEFAULT_ODB
from ebel.manager.orientdb.odb_meta import Graph

# Paths
VALIDATION_TEST_DIR = pathlib.Path(__file__).parent.absolute()
TEST_DATA_DIR = os.path.join(VALIDATION_TEST_DIR, "..", "data")

IMPORT_DIR = os.path.join(TEST_DATA_DIR, "import_tests")
TEST_BEL = os.path.join(IMPORT_DIR, "basic_import_test.bel")
TEST_JSON = TEST_BEL + ".json"

# Parameters for TEST database
USER = get_config_value(DEFAULT_ODB, "user")
PASSWORD = get_config_value(DEFAULT_ODB, "password")
DB_NAME = "ebel_test"
SERVER = "localhost"
PORT = "2424"
ROOT_PWD = get_config_value(DEFAULT_ODB, "root_password")

if ROOT_PWD is None:
    raise ValueError("Need root password to perform tests. Please add 'root_password' to configuration file")

# Configure client
config_params = {
    "db": DB_NAME,
    "user": USER,
    "password": PASSWORD,
    "server": SERVER,
    "port": PORT,
    "root_password": ROOT_PWD,
}

test_client = Bel(graph_config=config_params)



"""Defined variables used throughout the package."""
# -*- coding: utf-8 -*-

import os

THIS_DIR = os.path.dirname(__file__)
PROJECT_NAME = "ebel"

HOME = os.path.expanduser("~")
LIBRARY_NAME = PROJECT_NAME

# Path to folder
PROJECT_DIR = os.path.join(HOME, f".{PROJECT_NAME}")
if not os.path.exists(PROJECT_DIR):
    os.mkdir(PROJECT_DIR)

# Path to data folder
DATA_DIR = os.path.join(PROJECT_DIR, "data")
if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)

# Path to logs folder
LOG_DIR = os.path.join(PROJECT_DIR, "logs")
if not os.path.exists(LOG_DIR):
    os.mkdir(LOG_DIR)

# Default database name and location
DB_NAME = "{}.db".format(PROJECT_NAME)
DB_PATH = os.path.join(DATA_DIR, DB_NAME)

GRAMMAR_BEL_PATH = {
    "2": os.path.join(THIS_DIR, "grammar", "grammar_bel_2.bnf"),
    "2_1": os.path.join(THIS_DIR, "grammar", "grammar_bel_2_1.bnf"),
}

GRAMMAR_NS_ANNO_PATH = os.path.join(THIS_DIR, "grammar", "grammar_belns_belanno_1__2.bnf")
GRAMMAR_START_NS = "belns"
GRAMMAR_START_ANNO = "belanno"
GRAMMAR_START_LINE = "script_line_by_line"

# Variables
RID = "rid"

URL = "URL"
PATTERN = "PATTERN"
LIST = "LIST"
FILE = "FILE"
ALLOWED_TYPES = [URL, LIST, PATTERN, FILE]

DEFAULT_ODB = "DEFAULT_ODB"

# Protein Modifications
PMOD = {
    "pmod_ace": "Ac",
    "pmod_adr": "ADPRib",
    "pmod_add": "ADP-rybosylation",
    "pmod_far": "Farn",
    "pmod_ger": "Gerger",
    "pmod_gly": "Glyco",
    "pmod_hyd": "Hy",
    "pmod_isg": "ISG",
    "pmod_me0": "Me",
    "pmod_me1": "Me1",
    "pmod_mon": "monomethylation",
    "pmod_me2": "Me2",
    "pmod_me3": "Me3",
    "pmod_tri": "trimethylation",
    "pmod_myr": "Myr",
    "pmod_ned": "Nedd",
    "pmod_ngl": "NGlyco",
    "pmod_nit": "NO",
    "pmod_ogl": "OGlyco",
    "pmod_pal": "Palm",
    "pmod_pho": "Ph",
    "pmod_sul": "Sulf",
    "pmod_sup": "sulphation",
    "pmod_suh": "sulfonation",
    "pmod_sum": "Sumo",
    "pmod_suy": "Ub",
    "pmod_ubi": "ubiquitinylation",
    "pmod_u48": "UbK48",
    "pmod_u63": "UbK63",
    "pmod_ubm": "UbMono",
    "pmod_ubp": "UbPoly",
}

CYTOSCAPE_TEMPLATE = """{
  "format_version" : "1.0",
  "generated_by" : "cytoscape-3.6.0",
  "target_cytoscapejs_version" : "~2.1",
  "data" : {
    "shared_name" : "BEL_in_Cytoscape",
    "name" : "BEL_in_Cytoscape",
    "SUID" : 100000000,
    "__Annotations" : [ "" ],
    "networkMetadata" : "",
    "selected" : false
  },
  "elements" : {
    "nodes" : %s,
    "edges" : %s
  }
}"""

MD_FILE_HEADER = """Error report"""

SPECIES_NAMESPACE = {
    "HGNC": 9606,
    "MGI": 10090,
    "RGD": 10116,
    "FLYBASE": 7227,
}

# Edge types
# Causal
INCREASES = "increases"
DIRECTLY_INCREASES = "directly_increases"
CAUSAL_INCREASE = {INCREASES, DIRECTLY_INCREASES}

DECREASES = "decreases"
DIRECTLY_DECREASES = "directly_decreases"
CAUSAL_DECREASE = {DECREASES, DIRECTLY_DECREASES}

# Correlative
POSITIVE_CORRELATION = "positive_correlation"
NEGATIVE_CORRELATION = "negative_correlation"


# Terminal Colors
class TerminalFormatting:
    """Terminal style definitions."""

    RESET = "\033[0m"

    class Format:
        """Format style definitions."""

        BOLD = "\033[1m"
        ITALIC = "\033[3m"
        UNDERLINED = "\033[4m"
        STRIKETHROUGH = "\033[9m"

    class Fore:
        """Foreground style definitions."""

        __temp = "\033[{}m"
        __temp_b = "\033[1;{}m"
        RED = __temp.format(31)
        RED_BRIGHT = __temp_b.format(31)
        GREEN = __temp.format(32)
        GREEN_BRIGHT = __temp_b.format(32)
        YELLOW = __temp.format(33)
        YELLOW_BRIGHT = __temp_b.format(33)
        BLUE = __temp.format(34)
        BLUE_BRIGHT = __temp_b.format(34)
        PINK = __temp.format(35)
        PINK_BRIGHT = __temp_b.format(35)
        CYAN = __temp.format(36)
        CYAN_BRIGHT = __temp_b.format(36)
        WHITE = __temp_b.format(37)
        BLACK = __temp.format(30)
        GREY = __temp_b.format(30)
        GREY_BRIGHT = __temp_b.format(37)

    class Back:
        """Back style definitions."""

        __temp = "\033[{}m"
        NOTHING = __temp.format(40)
        RED = __temp.format(41)
        GREEN = __temp.format(42)
        YELLOW = __temp.format(43)
        BLUE = __temp.format(44)
        PINK = __temp.format(45)
        BLUE_LIGHT = __temp.format(46)
        GREY = __temp.format(47)

    TITLE = f"{Back.GREY}{Fore.BLUE}{Format.BOLD}{Format.UNDERLINED}"
    HEADER = f"{Fore.YELLOW}{Format.BOLD}{Format.UNDERLINED}"
    QUESTION = f"{Fore.GREEN}{Format.ITALIC}"
    DEFAULT_VALUE = f"{Fore.GREEN_BRIGHT}"



"""Base methods for API."""

from typing import Dict, List, Union

from pymysql.converters import escape_string
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from sqlalchemy_utils import create_database, database_exists

from ebel import Bel
from ebel.tools import _get_connection_string


class RDBMS:
    """Representation of the RDBMS."""

    engine = None
    session = None

    @staticmethod
    def get_engine():
        """Return the RDBMS engine."""
        if not RDBMS.engine:
            RDBMS.engine = create_engine(_get_connection_string())
        return RDBMS.engine

    @staticmethod
    def get_session():
        """Return the RDBMS session."""
        if not RDBMS.session:
            engine = RDBMS.get_engine()
            if not database_exists(engine.url):
                create_database(engine.url)

            RDBMS.session = Session(bind=engine, autocommit=True, autoflush=True)
        return RDBMS.session


class OdbRequest:
    """OrientDB class definition for interfacing with the ODB server."""

    def __init__(
        self,
        request_query_dict: Dict[str, Dict[str, Dict[str, Union[str, int, bool, float, None]]]],
    ):
        """Init method."""
        self.__request_query_dict = request_query_dict
        if self.validate():
            self.__odb_classes = request_query_dict
        else:
            raise TypeError("RequestQuery must be initialized with Dict[str, Dict[str, Dict[str, str]]]")

    @property
    def odb_classes(
        self,
    ) -> Dict[str, Dict[str, Dict[str, Union[str, int, bool, float, None]]]]:
        """Return ODB classes: structure: {'odb_class', {'odb_column': {'option':str, 'value': str}, ...}}}."""
        return self.__odb_classes

    def validate(self):
        """Check if request_query_dict is correct."""
        validated: bool = False
        if isinstance(self.__request_query_dict, Dict):
            for (
                odb_class,
                column_params,
            ) in self.__request_query_dict.items():  # k=ODB class, v=columns
                if isinstance(odb_class, str) and isinstance(column_params, Dict):
                    for column_name, value_option_dict in column_params.items():
                        if isinstance(value_option_dict, Dict):
                            keywords_exists = {"value", "option"}.issubset(set(value_option_dict.keys()))
                            value_types_ok = isinstance(
                                value_option_dict["value"],
                                (str, int, bool, float, type(None)),
                            )
                            if keywords_exists and value_types_ok:
                                validated = True
        return validated


class TabColSelectSql:
    """Formatter for OrientDb SQL columns Select part."""

    def __init__(self, odbclass_columns_dict, with_rids: bool = True, with_class: bool = True):
        """Init method."""
        self.odbclass_columns_dict = odbclass_columns_dict
        self.with_rids = with_rids
        self.with_class = with_class

    def __str__(self):
        """Define string method."""
        return self.get_sql()

    def __unicode__(self):
        """Unicode definition."""
        return self.__str__()

    def get_sql(self):
        """Retrieve the SQL query."""
        sql_list = []
        sql_class = "{odb_class}.@class.asString() as {odb_class}__class"
        sql_rid = "{odb_class}.@rid.asString() as {odb_class}__rid"
        for odb_class, columns in self.odbclass_columns_dict.items():
            if self.with_rids:
                sql_list.append(sql_rid.format(odb_class=odb_class))
            if self.with_class:
                sql_list.append(sql_class.format(odb_class=odb_class))
            for column in columns:
                sql_list.append(f"{odb_class}.{column} as {odb_class}__{column.replace('.', '_')}")
        return ", ".join(sql_list)


class ValueOption:
    """Class for defining the different values in a table."""

    def __init__(self, value: Union[str, None], option: str):
        """Init method."""
        self.value = value
        self.option = option

    def get_operator_value_sql_string(self, dialect: Union[str, None] = None) -> Union[str, None]:
        """Get "operator value" part of a WHERE SQL statement.

        Option is translated to dialect specific Operator like 'exact' to '='. Value will be escaped.

        Future implementations:
        Possible dialects are m (MySQL), o (OrientDB) or None. If dialect is None
        no further checks will be performed. Not implemented in the moment.

        Parameters
        ----------
        dialect: str
            SQL dialect can be set as m (=MySQL), o (=OrientDB).

        Returns
        -------
        str
            Operator-Value string.
        """
        if self.value:
            if isinstance(self.value, str):
                search_value = escape_string(self.value.strip(), "utf-8")

            else:
                search_value = self.value

            if self.option in ["=", "exact"]:
                operator = "="

            elif self.option == "regular expression":
                operator = "RLIKE"

            else:
                operator = "LIKE"

            option_choices = {
                "exact": f"'{self.value}'",
                "contains": f"'%{self.value}%'",
                "starts with": f"'{self.value}%'",
                "ends with": f"'{self.value}%'",
                "regular expression": f"'{self.value}'",
            }

            if self.option in option_choices:
                search_value = option_choices[self.option]

            elif self.option in [">", ">=", "<=", "<"]:
                operator = self.option

            elif operator == "LIKE":
                search_value = f"'%{self.value}%'"

            return f"{operator} {search_value}"


def get_sql_match(odb_request: OdbRequest, columns_dict: Dict[str, List[str]], match_template: str) -> str:
    """Return an OrientDB match string.

    Example:
        - request_query_dict = {'class_name1': {'column1': {'option': 'exact', 'value': 'value1'}}, ...}
        - columns_dict = {'class_name1': ['column1', 'column2'], ...}
        - match_template:
            match {{ class: class_name1, as: class_name1 {class_name1} }}
            .outE(){{ class: class_name2, as: class_name2 {class_name2} }}
            .inV(){{ class: class_name3, as:class_name3 {class_name3} }}
        - return_value: match_template:
            match {{ class: class_name1, as: class_name1, where:(column1='value1') }}
            .outE(){{ class: class_name2, as: class_name2 {class_name2} }}
            .inV(){{ class: class_name3, as:class_name3 {class_name3} }}

    Parameters
    ----------
    odb_request : OdbRequest
        Description of parameter `request_query_dict`.
    columns_dict : Dict[str, List[str]]
        Description of parameter `columns_dict`.
    match_template : str
        Description of parameter `sql_match_template`.

    Returns
    -------
    str
        Description of returned object.
    """
    where_dict = {}
    for odb_class in columns_dict.keys():
        if odb_request.odb_classes.get(odb_class):
            where_array = []
            for column, values in odb_request.odb_classes[odb_class].items():
                if values["value"]:
                    operator_value_string = ValueOption(**values).get_operator_value_sql_string()
                    if operator_value_string:
                        where_array.append(f"`{column}` {operator_value_string}")
            where_dict[odb_class] = ", where:(" + " and ".join(where_array) + ")" if where_array else ""

    sql_match = match_template.format(**where_dict)
    return sql_match


def get_odb_results(query_request: OdbRequest, class_column_dict: dict, sql_match_template: str):
    """Return OrientDB results for a given query."""
    query_get_dict = Bel().query_get_dict
    select_sql = TabColSelectSql(class_column_dict)
    sql_match = get_sql_match(query_request, class_column_dict, sql_match_template)

    query_string = f"""Select {select_sql}
        from
        ({sql_match}
        return {', '.join(class_column_dict.keys())})"""
    results = query_get_dict(query_string)

    return results


def get_odb_options(column: str, query_request: OdbRequest, class_column_dict, sql_match_template) -> List[str]:
    """Short summary.

    Returns
    -------
    List[str]
        Description of returned object.

    """
    client = Bel().client

    sql_match = get_sql_match(query_request, class_column_dict, sql_match_template)

    query_string = f"""Select distinct(option) as option
        from
        ({sql_match}
        return {column} as option) where option IS NOT NULL and option!='' order by option"""

    results = [str(x.oRecordData.get("option")) for x in client.command(query_string)]

    return results
